@@ -5,10 +5,26 @@
 //!
 //! Supports custom formatting in terms of degrees, minutes, and seconds, via the `Rad{32, 64}`.deg() method.
 
-use std::{fmt, marker::PhantomData, ops};
+use std::{
+    fmt,
+    marker::PhantomData,
+    ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Neg, Sub, SubAssign},
+};
 
 /// A floating-point number that serves as the backing value of an [`Angle`].
-pub trait Float: num_traits::Float {
+pub trait Float:
+    Copy
+    + PartialEq
+    + Add<Output = Self>
+    + AddAssign
+    + Sub<Output = Self>
+    + SubAssign
+    + Neg<Output = Self>
+    + Mul<Output = Self>
+    + MulAssign
+    + Div<Output = Self>
+    + DivAssign
+{
     /// Additive identity, 0.
     const ZERO: Self;
     /// Archimedes’ constant (π)
@@ -26,12 +42,36 @@ pub trait Float: num_traits::Float {
     /// The number 360.
     const THREE_SIXTY: Self;
 
+    /// Minimum finite value.
+    const MIN: Self;
+    /// Maximum finite value.
+    const MAX: Self;
+
+    /// Returns whether this value is finite (not infinity, not NaN).
+    fn is_finite(self) -> bool;
+    fn is_sign_positive(self) -> bool;
+    #[inline]
+    fn is_sign_negative(self) -> bool {
+        !self.is_sign_positive()
+    }
     /// Modulus operation.
     fn rem_euclid(self, _: Self) -> Self;
+    /// Absolute value.
+    fn abs(self) -> Self;
+    /// Truncates the fractional part from this value.
+    fn trunc(self) -> Self;
+
+    /// Type that implements [`std::cmp::Ord`], which this floating point type can be trivially converted to.
+    type Ord: Ord;
+    /// Converts this floating point number to a type that implements total ordering. Conversion should be trivial.  
+    ///
+    /// For insight on how to implement this, check out the README of https://github.com/notriddle/rust-float-ord.
+    /// Note that the implementation in that crate is slightly incorrect, as it will make 0.0 != -0.0
+    fn to_ord(self) -> Self::Ord;
 }
 
 macro_rules! impl_float {
-    ($f: ident) => {
+    ($f: ident, $i: ty) => {
         impl Float for $f {
             const ZERO: $f = 0.0;
             const PI: $f = std::$f::consts::PI;
@@ -41,17 +81,69 @@ macro_rules! impl_float {
             const NINETY: $f = 90.0;
             const ONE_EIGHTY: $f = 180.0;
             const THREE_SIXTY: $f = 360.0;
+            const MIN: $f = $f::MIN;
+            const MAX: $f = $f::MAX;
 
+            #[inline]
+            fn is_finite(self) -> bool {
+                <$f>::is_finite(self)
+            }
+            #[inline]
+            fn is_sign_positive(self) -> bool {
+                <$f>::is_sign_positive(self)
+            }
+            #[inline]
+            fn is_sign_negative(self) -> bool {
+                <$f>::is_sign_negative(self)
+            }
             #[inline]
             fn rem_euclid(self, rhs: Self) -> Self {
                 <$f>::rem_euclid(self, rhs)
             }
+            #[inline]
+            fn abs(self) -> Self {
+                <$f>::abs(self)
+            }
+            #[inline]
+            fn trunc(self) -> Self {
+                <$f>::trunc(self)
+            }
+
+            type Ord = $i;
+            #[inline]
+            fn to_ord(self) -> Self::Ord {
+                // assert that the types have equal sizes.
+                const _ASSERT: [(); std::mem::size_of::<$f>()] = [(); std::mem::size_of::<$i>()];
+
+                const MSB: $i = 1 << (std::mem::size_of::<$f>() * 8 - 1);
+                let bits = self.to_bits();
+                if bits & MSB == 0 {
+                    // if it's positive, flip the most significant bit.
+                    bits | MSB
+                } else {
+                    // if its negative zero, pretend that its positive zero.
+                    if bits << 1 == 0 {
+                        return zero();
+
+                        // Benchmarking shows that marking as cold provides a slight performance boost.
+                        // This is because -0 is a special case.
+                        #[cold]
+                        fn zero() -> $i {
+                            MSB // this is the result of flipping the most significant bit of +0
+                        }
+                    }
+                    // if it's any other negative number, flip every bit
+                    else {
+                        !bits
+                    }
+                }
+            }
         }
     };
 }
 
-impl_float!(f32);
-impl_float!(f64);
+impl_float!(f32, u32);
+impl_float!(f64, u64);
 
 /// A unit of measurement for an [`Angle`].
 pub trait Unit<F: Float>: Sized {
@@ -88,13 +180,20 @@ impl<F: Float, U: Unit<F>> Copy for Angle<F, U> {}
 impl<F: Float, U: Unit<F>> PartialEq for Angle<F, U> {
     #[inline]
     fn eq(&self, rhs: &Self) -> bool {
-        self.0 == rhs.0
+        self.0.to_ord() == rhs.0.to_ord()
     }
 }
+impl<F: Float, U: Unit<F>> Eq for Angle<F, U> {}
 impl<F: Float, U: Unit<F>> PartialOrd for Angle<F, U> {
     #[inline]
     fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
-        self.0.partial_cmp(&other.0)
+        Some(self.cmp(other))
+    }
+}
+impl<F: Float, U: Unit<F>> Ord for Angle<F, U> {
+    #[inline]
+    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
+        self.0.to_ord().cmp(&other.0.to_ord())
     }
 }
 
@@ -120,6 +219,10 @@ impl<F: Float, U: Unit<F>> Angle<F, U> {
     pub const HALF_TURN: Self = Self(U::HALF_TURN, PhantomData);
     /// Full turn around a circle. Equal to 2π radians or 360°.
     pub const FULL_TURN: Self = Self(U::FULL_TURN, PhantomData);
+    /// Minimum finite angle.
+    pub const MIN: Self = Self(F::MIN, PhantomData);
+    /// Maximum finite angle.
+    pub const MAX: Self = Self(F::MAX, PhantomData);
 
     /// Create a new angle from a raw value.
     /// # Panics
@@ -141,40 +244,41 @@ impl<F: Float, U: Unit<F>> Angle<F, U> {
     }
     /// Returns the magnitude (absolute value) of this angle.
     #[inline]
-    pub fn mag(self) -> F {
-        self.0.abs()
+    pub fn mag(mut self) -> Self {
+        self.0 = self.0.abs();
+        self
     }
 }
 
-impl<F: Float, U: Unit<F>> ops::Add for Angle<F, U> {
+impl<F: Float, U: Unit<F>> Add for Angle<F, U> {
     type Output = Self;
     #[inline]
     fn add(self, rhs: Self) -> Self {
         Self::new(self.0 + rhs.0)
     }
 }
-impl<F: Float + ops::AddAssign, U: Unit<F>> ops::AddAssign for Angle<F, U> {
+impl<F: Float, U: Unit<F>> AddAssign for Angle<F, U> {
     #[inline]
     fn add_assign(&mut self, rhs: Self) {
         self.0 += rhs.0;
         debug_assert!(self.0.is_finite());
     }
 }
-impl<F: Float, U: Unit<F>> ops::Sub for Angle<F, U> {
+impl<F: Float, U: Unit<F>> Sub for Angle<F, U> {
     type Output = Self;
     #[inline]
     fn sub(self, rhs: Self) -> Self {
         Self::new(self.0 - rhs.0)
     }
 }
-impl<F: Float + ops::SubAssign, U: Unit<F>> ops::SubAssign for Angle<F, U> {
+impl<F: Float, U: Unit<F>> SubAssign for Angle<F, U> {
     #[inline]
     fn sub_assign(&mut self, rhs: Self) {
         self.0 -= rhs.0;
         debug_assert!(self.0.is_finite());
     }
 }
-impl<F: Float, U: Unit<F>> ops::Neg for Angle<F, U> {
+impl<F: Float, U: Unit<F>> Neg for Angle<F, U> {
     type Output = Self;
     #[inline]
     fn neg(self) -> Self {
@@ -182,28 +286,28 @@ impl<F: Float, U: Unit<F>> ops::Neg for Angle<F, U> {
     }
 }
 
-impl<F: Float, U: Unit<F>> ops::Mul<F> for Angle<F, U> {
+impl<F: Float, U: Unit<F>> Mul<F> for Angle<F, U> {
     type Output = Self;
     #[inline]
     fn mul(self, rhs: F) -> Self {
         Self::new(self.0 * rhs)
     }
 }
-impl<F: Float + ops::MulAssign, U: Unit<F>> ops::MulAssign<F> for Angle<F, U> {
+impl<F: Float, U: Unit<F>> MulAssign<F> for Angle<F, U> {
     #[inline]
     fn mul_assign(&mut self, rhs: F) {
         self.0 *= rhs;
         debug_assert!(self.0.is_finite());
     }
 }
-impl<F: Float, U: Unit<F>> ops::Div<F> for Angle<F, U> {
+impl<F: Float, U: Unit<F>> Div<F> for Angle<F, U> {
     type Output = Self;
     #[inline]
     fn div(self, rhs: F) -> Self {
         Self::new(self.0 / rhs)
     }
 }
-impl<F: Float + ops::DivAssign, U: Unit<F>> ops::DivAssign<F> for Angle<F, U> {
+impl<F: Float, U: Unit<F>> DivAssign<F> for Angle<F, U> {
     #[inline]
     fn div_assign(&mut self, rhs: F) {
         self.0 /= rhs;
@@ -328,15 +432,10 @@ impl<F: Float, U: Unit<F>> Copy for Wrap<F, U> {}
 impl<F: Float, U: Unit<F>> PartialEq for Wrap<F, U> {
     #[inline]
     fn eq(&self, rhs: &Self) -> bool {
-        self.0 .0 == rhs.0 .0
-    }
-}
-impl<F: Float, U: Unit<F>> PartialOrd for Wrap<F, U> {
-    #[inline]
-    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
-        self.0 .0.partial_cmp(&other.0 .0)
+        self.0 == rhs.0
     }
 }
+impl<F: Float, U: Unit<F>> Eq for Wrap<F, U> {}
 
 impl<F: Float, U: Unit<F>> Wrap<F, U> {
     /// Zero angle, additive identity.
@@ -359,6 +458,16 @@ impl<F: Float, U: Unit<F>> Wrap<F, U> {
     pub fn val_raw(self) -> F {
         self.0 .0
     }
+    /// Gets the inner representation of this angle.
+    #[inline]
+    pub fn inner(self) -> Angle<F, U> {
+        self.0
+    }
+    /// Returns the magnitude (absolute value) of this angle.
+    #[inline]
+    pub fn mag(self) -> Angle<F, U> {
+        self.0.mag()
+    }
 }
 
 impl<F: Float> Wrap<F, Radians> {
@@ -383,33 +492,33 @@ impl<F: Float, U: Unit<F>> From<Wrap<F, U>> for Angle<F, U> {
     }
 }
 
-impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> ops::Add<Rhs> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> Add<Rhs> for Wrap<F, U> {
     type Output = Self;
     #[inline]
     fn add(self, rhs: Rhs) -> Self {
         Self::wrap(self.val_raw() + rhs.into().0)
     }
 }
-impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> ops::AddAssign<Rhs> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> AddAssign<Rhs> for Wrap<F, U> {
     #[inline]
     fn add_assign(&mut self, rhs: Rhs) {
         *self = *self + rhs
     }
 }
-impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> ops::Sub<Rhs> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> Sub<Rhs> for Wrap<F, U> {
     type Output = Self;
     #[inline]
     fn sub(self, rhs: Rhs) -> Self {
         Self::wrap(self.val_raw() + rhs.into().0)
     }
 }
-impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> ops::SubAssign<Rhs> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>, Rhs: Into<Angle<F, U>>> SubAssign<Rhs> for Wrap<F, U> {
     #[inline]
     fn sub_assign(&mut self, rhs: Rhs) {
         *self = *self - rhs;
     }
 }
-impl<F: Float, U: Unit<F>> ops::Neg for Wrap<F, U> {
+impl<F: Float, U: Unit<F>> Neg for Wrap<F, U> {
     type Output = Self;
     #[inline]
     fn neg(self) -> Self {
@@ -417,27 +526,27 @@ impl<F: Float, U: Unit<F>> ops::Neg for Wrap<F, U> {
     }
 }
 
-impl<F: Float, U: Unit<F>> ops::Mul<F> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>> Mul<F> for Wrap<F, U> {
     type Output = Self;
     #[inline]
     fn mul(self, rhs: F) -> Self {
         Self::wrap(self.val_raw() * rhs)
     }
 }
-impl<F: Float, U: Unit<F>> ops::MulAssign<F> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>> MulAssign<F> for Wrap<F, U> {
     #[inline]
     fn mul_assign(&mut self, rhs: F) {
         *self = *self * rhs;
     }
 }
-impl<F: Float, U: Unit<F>> ops::Div<F> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>> Div<F> for Wrap<F, U> {
     type Output = Self;
     #[inline]
     fn div(self, rhs: F) -> Self {
         Self::wrap(self.val_raw() / rhs)
     }
 }
-impl<F: Float, U: Unit<F>> ops::DivAssign<F> for Wrap<F, U> {
+impl<F: Float, U: Unit<F>> DivAssign<F> for Wrap<F, U> {
     #[inline]
     fn div_assign(&mut self, rhs: F) {
         *self = *self / rhs
@@ -459,21 +568,37 @@ impl<F: Float + fmt::Display, U: Unit<F>> fmt::Display for Wrap<F, U> {
 
 /// A 32-bit angle measured in radians.
 ///
+/// This type is guaranteed to be finite (in debug mode). As such, it implements total equality and ordering.
+///
 /// When formatting with [`Display`](fmt::Display), it will be shown as a multiple of π.
 pub type Rad32 = Rad<f32>;
 /// A 64-bit angle measured in radians.
 ///
+/// This type is guaranteed to be finite (in debug mode). As such, it implements total equality and ordering.
+///
 /// When formatting with [`Display`](fmt::Display), it will be shown as a multiple of π.
 pub type Rad64 = Rad<f64>;
 
 /// A 32-bit angle measured in degrees.
+///
+/// This type is guaranteed to be finite (in debug mode). As such, it implements total equality and ordering.
 pub type Deg32 = Deg<f32>;
 /// A 64-bit angle measured in degrees.
+///
+/// This type is guaranteed to be finite (in debug mode). As such, it implements total equality and ordering.
 pub type Deg64 = Deg<f64>;
 
 /// A 32-bit angle measured in radians that wraps between -π and +π.
+///
+/// This type is guaranteed to be finite (in debug mode). As such, it implements total equality.  
+/// It intentionally does not implement any ordering, as the wrapping behavior of this type may
+/// cause unexpected ordering. If you need ordering, use the method `inner` or `mag`.
 pub type Wrap32 = Wrap<f32, Radians>;
 /// A 64-bit angle measured in radians that wraps between -π and +π.
+///
+/// This type is guaranteed to be finite (in debug mode). As such, it implements total equality.  
+/// It intentionally does not implement any ordering, as the wrapping behavior of this type may
+/// cause unexpected ordering. If you need ordering, use the method `inner` or `mag`.
 pub type Wrap64 = Wrap<f64, Radians>;
 
 #[cfg(test)]
@@ -506,6 +631,25 @@ mod tests {
         assert_epsilon!(Wrap64::FULL_TURN.val(), 2.0 * PI);
     }
 
+    #[test]
+    fn ord() {
+        assert!(Rad64::MIN < -Rad64::FULL_TURN);
+        assert!(-Rad64::FULL_TURN < -Rad64::QUARTER_TURN);
+        assert!(-Rad64::QUARTER_TURN < -Rad64::ZERO);
+        assert!(-Rad64::ZERO == Rad64::ZERO);
+        assert!(Rad64::ZERO < Rad64::QUARTER_TURN);
+        assert!(Rad64::QUARTER_TURN < Rad64::FULL_TURN);
+        assert!(Rad64::FULL_TURN < Rad64::MAX);
+
+        assert!(Rad64::MAX > Rad64::FULL_TURN);
+        assert!(Rad64::FULL_TURN > Rad64::QUARTER_TURN);
+        assert!(Rad64::QUARTER_TURN > Rad64::ZERO);
+        assert!(Rad64::ZERO == -Rad64::ZERO);
+        assert!(-Rad64::ZERO > -Rad64::QUARTER_TURN);
+        assert!(-Rad64::QUARTER_TURN > -Rad64::FULL_TURN);
+        assert!(-Rad64::FULL_TURN > Rad64::MIN);
+    }
+
     #[test]
     fn rad_ops() {
         let sum = Rad64::HALF_TURN + Rad64::HALF_TURN;
